@@ -8,9 +8,8 @@ use std::collections::BTreeSet;
 use std::iter;
 
 use derive_builder::Builder;
-use itertools::Itertools;
 
-use crate::api::common::NameOrId;
+use crate::api::common::{CommaSeparatedList, NameOrId};
 use crate::api::endpoint_prelude::*;
 
 #[derive(Debug, Clone)]
@@ -36,6 +35,28 @@ impl Assignee {
     }
 }
 
+/// Parameters for setting the reviewer(s) of a merge request.
+#[derive(Debug, Clone)]
+pub(crate) enum Reviewer {
+    /// Unset all reviewers.
+    Unassigned,
+    /// A set of reviewers.
+    Ids(BTreeSet<u64>),
+}
+
+impl Reviewer {
+    pub(crate) fn add_params<'a>(&'a self, params: &mut FormParams<'a>) {
+        match self {
+            Reviewer::Unassigned => {
+                params.push("reviewer_ids", "0");
+            },
+            Reviewer::Ids(ids) => {
+                params.extend(ids.iter().map(|&id| ("reviewer_ids[]", id)));
+            },
+        }
+    }
+}
+
 /// Create a new merge request on project.
 #[derive(Debug, Builder)]
 #[builder(setter(strip_option))]
@@ -56,6 +77,8 @@ pub struct CreateMergeRequest<'a> {
     /// The assignee of the merge request.
     #[builder(setter(name = "_assignee"), default, private)]
     assignee: Option<Assignee>,
+    #[builder(setter(name = "_reviewer"), default, private)]
+    reviewer: Option<Reviewer>,
     /// The description of the merge request.
     #[builder(setter(into), default)]
     description: Option<Cow<'a, str>>,
@@ -64,10 +87,15 @@ pub struct CreateMergeRequest<'a> {
     target_project_id: Option<u64>,
     /// Labels to add to the merge request.
     #[builder(setter(name = "_labels"), default, private)]
-    labels: BTreeSet<Cow<'a, str>>,
+    labels: Option<CommaSeparatedList<Cow<'a, str>>>,
     /// The ID of the milestone to add the merge request to.
     #[builder(default)]
     milestone_id: Option<u64>,
+    /// How many approvals are required before merging will be allowed.
+    ///
+    /// Note that this must be more than the project limit (if present).
+    #[builder(default)]
+    approvals_before_merge: Option<u64>,
     /// Whether to remove the source branch once merged or not.
     #[builder(default)]
     remove_source_branch: Option<bool>,
@@ -135,14 +163,50 @@ impl<'a> CreateMergeRequestBuilder<'a> {
         self
     }
 
+    /// Filter merge requests without a reviewer.
+    pub fn without_reviewer(&mut self) -> &mut Self {
+        self.reviewer = Some(Some(Reviewer::Unassigned));
+        self
+    }
+
+    /// Filter merge requests reviewed by a user (by ID).
+    pub fn reviewer(&mut self, reviewer: u64) -> &mut Self {
+        let reviewer = match self.reviewer.take() {
+            Some(Some(Reviewer::Ids(mut set))) => {
+                set.insert(reviewer);
+                Reviewer::Ids(set)
+            },
+            _ => Reviewer::Ids(iter::once(reviewer).collect()),
+        };
+        self.reviewer = Some(Some(reviewer));
+        self
+    }
+
+    /// Filter merge requests reviewed by users (by ID).
+    pub fn reviewers<I>(&mut self, iter: I) -> &mut Self
+    where
+        I: Iterator<Item = u64>,
+    {
+        let reviewer = match self.reviewer.take() {
+            Some(Some(Reviewer::Ids(mut set))) => {
+                set.extend(iter);
+                Reviewer::Ids(set)
+            },
+            _ => Reviewer::Ids(iter.collect()),
+        };
+        self.reviewer = Some(Some(reviewer));
+        self
+    }
+
     /// Add a label.
     pub fn label<L>(&mut self, label: L) -> &mut Self
     where
         L: Into<Cow<'a, str>>,
     {
         self.labels
-            .get_or_insert_with(BTreeSet::new)
-            .insert(label.into());
+            .get_or_insert(None)
+            .get_or_insert_with(CommaSeparatedList::new)
+            .push(label.into());
         self
     }
 
@@ -153,7 +217,8 @@ impl<'a> CreateMergeRequestBuilder<'a> {
         L: Into<Cow<'a, str>>,
     {
         self.labels
-            .get_or_insert_with(BTreeSet::new)
+            .get_or_insert(None)
+            .get_or_insert_with(CommaSeparatedList::new)
             .extend(iter.map(Into::into));
         self
     }
@@ -178,16 +243,18 @@ impl<'a> Endpoint for CreateMergeRequest<'a> {
             .push_opt("description", self.description.as_ref())
             .push_opt("target_project_id", self.target_project_id)
             .push_opt("milestone_id", self.milestone_id)
+            .push_opt("labels", self.labels.as_ref())
+            .push_opt("approvals_before_merge", self.approvals_before_merge)
             .push_opt("remove_source_branch", self.remove_source_branch)
             .push_opt("allow_collaboration", self.allow_collaboration)
             .push_opt("squash", self.squash);
 
-        if !self.labels.is_empty() {
-            params.push("labels", format!("{}", self.labels.iter().format(",")));
-        }
         if let Some(assignee) = self.assignee.as_ref() {
             assignee.add_params(&mut params);
         }
+        if let Some(reviewer) = self.reviewer.as_ref() {
+            reviewer.add_params(&mut params);
+        }
 
         #[allow(deprecated)]
         {
@@ -202,14 +269,16 @@ impl<'a> Endpoint for CreateMergeRequest<'a> {
 mod tests {
     use http::Method;
 
-    use crate::api::projects::merge_requests::CreateMergeRequest;
+    use crate::api::projects::merge_requests::{
+        CreateMergeRequest, CreateMergeRequestBuilderError,
+    };
     use crate::api::{self, Query};
     use crate::test::client::{ExpectedUrl, SingleTestClient};
 
     #[test]
     fn project_source_branch_target_branch_and_title_are_necessary() {
         let err = CreateMergeRequest::builder().build().unwrap_err();
-        assert_eq!(err, "`project` must be initialized");
+        crate::test::assert_missing_field!(err, CreateMergeRequestBuilderError, "project");
     }
 
     #[test]
@@ -220,7 +289,7 @@ mod tests {
             .title("title")
             .build()
             .unwrap_err();
-        assert_eq!(err, "`project` must be initialized");
+        crate::test::assert_missing_field!(err, CreateMergeRequestBuilderError, "project");
     }
 
     #[test]
@@ -231,7 +300,7 @@ mod tests {
             .title("title")
             .build()
             .unwrap_err();
-        assert_eq!(err, "`source_branch` must be initialized");
+        crate::test::assert_missing_field!(err, CreateMergeRequestBuilderError, "source_branch");
     }
 
     #[test]
@@ -242,7 +311,7 @@ mod tests {
             .title("title")
             .build()
             .unwrap_err();
-        assert_eq!(err, "`target_branch` must be initialized");
+        crate::test::assert_missing_field!(err, CreateMergeRequestBuilderError, "target_branch");
     }
 
     #[test]
@@ -253,7 +322,7 @@ mod tests {
             .target_branch("target")
             .build()
             .unwrap_err();
-        assert_eq!(err, "`title` must be initialized");
+        crate::test::assert_missing_field!(err, CreateMergeRequestBuilderError, "title");
     }
 
     #[test]
@@ -375,6 +444,89 @@ mod tests {
         api::ignore(endpoint).query(&client).unwrap();
     }
 
+    #[test]
+    fn endpoint_unreviewed() {
+        let endpoint = ExpectedUrl::builder()
+            .method(Method::POST)
+            .endpoint("projects/simple%2Fproject/merge_requests")
+            .content_type("application/x-www-form-urlencoded")
+            .body_str(concat!(
+                "source_branch=source%2Fbranch",
+                "&target_branch=target%2Fbranch",
+                "&title=title",
+                "&reviewer_ids=0",
+            ))
+            .build()
+            .unwrap();
+        let client = SingleTestClient::new_raw(endpoint, "");
+
+        let endpoint = CreateMergeRequest::builder()
+            .project("simple/project")
+            .source_branch("source/branch")
+            .target_branch("target/branch")
+            .title("title")
+            .without_reviewer()
+            .build()
+            .unwrap();
+        api::ignore(endpoint).query(&client).unwrap();
+    }
+
+    #[test]
+    fn endpoint_reviewer() {
+        let endpoint = ExpectedUrl::builder()
+            .method(Method::POST)
+            .endpoint("projects/simple%2Fproject/merge_requests")
+            .content_type("application/x-www-form-urlencoded")
+            .body_str(concat!(
+                "source_branch=source%2Fbranch",
+                "&target_branch=target%2Fbranch",
+                "&title=title",
+                "&reviewer_ids%5B%5D=1",
+            ))
+            .build()
+            .unwrap();
+        let client = SingleTestClient::new_raw(endpoint, "");
+
+        let endpoint = CreateMergeRequest::builder()
+            .project("simple/project")
+            .source_branch("source/branch")
+            .target_branch("target/branch")
+            .title("title")
+            .reviewer(1)
+            .build()
+            .unwrap();
+        api::ignore(endpoint).query(&client).unwrap();
+    }
+
+    #[test]
+    fn endpoint_reviewers() {
+        let endpoint = ExpectedUrl::builder()
+            .method(Method::POST)
+            .endpoint("projects/simple%2Fproject/merge_requests")
+            .content_type("application/x-www-form-urlencoded")
+            .body_str(concat!(
+                "source_branch=source%2Fbranch",
+                "&target_branch=target%2Fbranch",
+                "&title=title",
+                "&reviewer_ids%5B%5D=1",
+                "&reviewer_ids%5B%5D=2",
+            ))
+            .build()
+            .unwrap();
+        let client = SingleTestClient::new_raw(endpoint, "");
+
+        let endpoint = CreateMergeRequest::builder()
+            .project("simple/project")
+            .source_branch("source/branch")
+            .target_branch("target/branch")
+            .title("title")
+            .reviewer(1)
+            .reviewers([1, 2].iter().copied())
+            .build()
+            .unwrap();
+        api::ignore(endpoint).query(&client).unwrap();
+    }
+
     #[test]
     fn endpoint_description() {
         let endpoint = ExpectedUrl::builder()
@@ -484,6 +636,33 @@ mod tests {
         api::ignore(endpoint).query(&client).unwrap();
     }
 
+    #[test]
+    fn endpoint_approvals_before_merge() {
+        let endpoint = ExpectedUrl::builder()
+            .method(Method::POST)
+            .endpoint("projects/simple%2Fproject/merge_requests")
+            .content_type("application/x-www-form-urlencoded")
+            .body_str(concat!(
+                "source_branch=source%2Fbranch",
+                "&target_branch=target%2Fbranch",
+                "&title=title",
+                "&approvals_before_merge=2",
+            ))
+            .build()
+            .unwrap();
+        let client = SingleTestClient::new_raw(endpoint, "");
+
+        let endpoint = CreateMergeRequest::builder()
+            .project("simple/project")
+            .source_branch("source/branch")
+            .target_branch("target/branch")
+            .title("title")
+            .approvals_before_merge(2)
+            .build()
+            .unwrap();
+        api::ignore(endpoint).query(&client).unwrap();
+    }
+
     #[test]
     fn endpoint_remove_source_branch() {
         let endpoint = ExpectedUrl::builder()
